#!/usr/bin/env python

from datetime import date, datetime
from typing import Optional, Union

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

try:  # pragma: no cover
    from secrets import SystemRandom  # (Python >= 3.6) type: ignore
except ImportError:  # pragma: no cover
    from random import SystemRandom


__all__ = [
    "AES_BLOCK_SIZE",
    "AES_KEY_SIZE",
    "Bytes",
    "BytesStr",
    "random",
    "Bytearray",
    "aes_ecb_encrypt",
    "aes_ecb_decrypt",
    "xor_block",
    "cbc_hash",
    "fromisoformat",
]

AES_BLOCK_SIZE = 16
AES_KEY_SIZE = 16

Bytes = Union[bytes, bytearray, "Bytearray"]
BytesStr = Union[bytes, bytearray, str, "Bytearray"]


random = SystemRandom()


class Bytearray(bytearray):
    def arrayset(self, c: int, n: int, dest_offset: int = 0) -> None:
        self[dest_offset : dest_offset + n] = [c] * n

    def arraycpy(self, src: BytesStr, n: Optional[int] = None, dest_offset: int = 0) -> None:
        if isinstance(src, str):
            src = bytes(src, "ascii")
        if n is None:
            n = len(src)
        n = min(n, len(self) - dest_offset, len(src))
        self[dest_offset : dest_offset + n] = src[0:n]


def aes_ecb_encrypt(key: Bytes, data: Bytes) -> bytes:
    """
    Encrypt data with the key using AES-128 ECB
    """
    cipher = Cipher(algorithms.AES(bytes(key)), modes.ECB())
    encryptor = cipher.encryptor()  # type: ignore
    return encryptor.update(bytes(data))  # type: ignore


def aes_ecb_decrypt(key: Bytes, data: Bytes) -> bytes:
    """
    Decrypt data with the key using AES-128 ECB
    """
    cipher = Cipher(algorithms.AES(bytes(key)), modes.ECB())
    decryptor = cipher.decryptor()  # type: ignore
    return decryptor.update(bytes(data))  # type: ignore


def xor_block(a: Bytes, b: Bytes) -> bytes:
    return bytes(a[i] ^ (b[i] if i < len(b) else 0) for i in range(0, len(a)))


def cbc_hash(key: Bytes, iv: Bytes, data: Bytes) -> bytes:
    """
    Calculate cipher block chaining message authentication code
    """
    result = bytes(iv)
    while len(data) > 0:
        result = aes_ecb_encrypt(key, xor_block(result, data))
        data = data[AES_BLOCK_SIZE:]
    return bytes(result)


def fromisoformat(dt: str) -> date:
    """
    Convert a YYYY-MM-DD string into a date object
    """
    return datetime.strptime(dt, "%Y-%m-%d").date()



#!/usr/bin/env python

import math
import os
import os.path
from datetime import date
from typing import Optional, Union

from .exceptions import (
    InvalidSeed,
    InvalidSerial,
    InvalidSignature,
    InvalidToken,
    ParseException,
)
from .token import SERIAL_LENGTH, AbstractTokenFile, Token
from .utils import AES_KEY_SIZE, Bytearray, Bytes, aes_ecb_decrypt, aes_ecb_encrypt

__all__ = ["DEFAULT_STOKEN_FILENAME", "StokenFile"]

DEFAULT_STOKEN_FILENAME = "~/.stokenrc"
FL_128BIT = 1 << 14
FL_PASSPROT = 1 << 13
FL_SNPROT = 1 << 12
FL_APPSEEDS = 1 << 11
FL_FEAT4 = 1 << 10
FL_TIMESEEDS = 1 << 9
FLD_DIGIT_SHIFT = 6
FLD_DIGIT_MASK = 0x07 << FLD_DIGIT_SHIFT
FLD_PINMODE_SHIFT = 3
FLD_PINMODE_MASK = 0x03 << FLD_PINMODE_SHIFT
FLD_NUMSECONDS_SHIFT = 0
FLD_NUMSECONDS_MASK = 0x03 << FLD_NUMSECONDS_SHIFT
TOKEN_BITS_PER_CHAR = 3
STOKEN_MAGIC = bytes([0xD8, 0xF5, 0x32, 0x53, 0x82, 0x89])
VER_LENGTH = 1
CHECKSUM_BITS = 15
CHECKSUM_LENGTH = int(CHECKSUM_BITS / TOKEN_BITS_PER_CHAR)
BINENC_BITS = 189
BINENC_OFS = VER_LENGTH + SERIAL_LENGTH
MIN_TOKEN_BITS = 189
MAX_TOKEN_BITS = 255
MAX_TOKEN_LENGTH = int(MAX_TOKEN_BITS / TOKEN_BITS_PER_CHAR)
MIN_TOKEN_LENGTH = int((MIN_TOKEN_BITS / TOKEN_BITS_PER_CHAR) + SERIAL_LENGTH + VER_LENGTH + CHECKSUM_LENGTH)
SECURID_EPOCH = 730120  # 2000/01/01 proleptic Gregorian ordinal


class StokenFile(AbstractTokenFile):
    """
    Handler for stokenrc file format
    """

    filename: Optional[str]
    data: Optional[bytes]
    token: Token

    def __init__(
        self,
        filename: Optional[str] = DEFAULT_STOKEN_FILENAME,
        data: Union[Optional[bytes], Optional[bytearray], Optional[str]] = None,
        token: Optional[Token] = None,
    ) -> None:
        """
        :param filename: stokenrc file path
        :param data: token as string in stokenrc format
        :param token: Token instance
        """
        if token is not None:
            self.filename = None
            self.token = token
        elif data is not None:
            if isinstance(data, str):
                data = bytes(data, "ascii")
            self.filename = None
            self.data = data
            self.token = self.v2_decode_token(self.data)
        elif filename is not None:
            self.filename = os.path.expanduser(filename)
            self.data = self.parse_file(self.filename)
            self.token = self.v2_decode_token(self.data)
            self.token.pin = self.parse_file_pin(self.filename)

    @classmethod
    def parse_file(cls, filename: str) -> bytes:
        """
        Parse stokenrc file, return token as string

        :param filename: stokenrc file path
        """
        with open(filename, "rb") as f:
            for line in f.readlines():
                line = line.strip()
                if b" " in line:
                    k, v = line.split(b" ", 1)
                    if k == b"token":
                        return v
        raise ParseException("Error parsing {}".format(filename))

    @classmethod
    def parse_file_pin(cls, filename: str) -> int:
        """
        Parse stokenrc file, return pin as int or 0 if not found

        :param filename: stokenrc file path
        """
        with open(filename, "rb") as f:
            for line in f.readlines():
                line = line.strip()
                if b" " in line:
                    k, v = line.split(b" ", 1)
                    if k == b"pin":
                        return int(v)
        return 0

    @classmethod
    def v2_decode_token(cls, data: Bytes) -> Token:
        if len(data) < MIN_TOKEN_LENGTH or len(data) > MAX_TOKEN_LENGTH:
            raise InvalidToken("Invalid token length")
        cls._verify_checksum(data)
        # version = data[0] - ord('0')
        d = cls._numinput_to_bits(data, BINENC_BITS, offset=BINENC_OFS)
        enc_seed = d[0:AES_KEY_SIZE]
        flags = cls._get_bits(d, 128, 16)
        seed_hash = cls._get_bits(d, 159, 15)
        exp_date = date.fromordinal(cls._get_bits(d, 144, 14) + SECURID_EPOCH)
        serial = str(data[VER_LENGTH : VER_LENGTH + SERIAL_LENGTH], "ascii")
        interval = 60 if ((flags & FLD_NUMSECONDS_MASK) >> FLD_NUMSECONDS_SHIFT) else 30
        digits = ((flags & FLD_DIGIT_MASK) >> FLD_DIGIT_SHIFT) + 1
        seed = cls.v2_decrypt_seed(enc_seed, seed_hash)
        return Token(
            serial=serial,
            seed=seed,
            interval=interval,
            digits=digits,
            exp_date=exp_date,
        )

    def v2_encode_token(self) -> None:
        if not self.token.seed:
            raise InvalidSeed("Missing seed")
        if not self.token.serial:
            raise InvalidSerial("Missing serial")
        if len(self.token.serial) != SERIAL_LENGTH:
            raise InvalidSerial("Serial length != {}".format(SERIAL_LENGTH))
        flags = FL_TIMESEEDS | FL_128BIT
        flags |= (self.token.digits - 1 << FLD_DIGIT_SHIFT) & FLD_DIGIT_MASK
        flags |= (1 << FLD_NUMSECONDS_SHIFT) if self.token.interval == 60 else 0

        key_hash = self._securid_mac(STOKEN_MAGIC)
        enc_seed = aes_ecb_encrypt(key_hash, self.token.seed)
        seed_hash = self._short_hash(self._securid_mac(self.token.seed))

        d = Bytearray(int(MAX_TOKEN_BITS / 8 + 2))
        d.arraycpy(enc_seed, n=AES_KEY_SIZE, dest_offset=0)
        self._set_bits(d, 128, 16, flags)
        self._set_bits(d, 159, 15, seed_hash)
        self._set_bits(
            d,
            144,
            14,
            (self.token.exp_date.toordinal() - SECURID_EPOCH) if self.token.exp_date else 0,
        )

        data = Bytearray(81)
        data.arraycpy(b"2", dest_offset=0)  # version
        data.arraycpy(self._bits_to_numoutput(d, BINENC_BITS), dest_offset=BINENC_OFS)
        data.arraycpy(self.token.serial, n=SERIAL_LENGTH, dest_offset=VER_LENGTH)

        d = Bytearray(int(MAX_TOKEN_BITS / 8 + 2))
        computed_mac = self._securid_shortmac(data[: len(data) - CHECKSUM_LENGTH])
        self._set_bits(d, 0, 15, computed_mac)
        t = self._bits_to_numoutput(d, 15)
        data.arraycpy(t, dest_offset=len(data) - CHECKSUM_LENGTH)

        self._verify_checksum(data)
        self.data = bytes(data)

    @classmethod
    def _verify_checksum(cls, data: Bytes) -> None:
        d = cls._numinput_to_bits(data, 15, offset=len(data) - CHECKSUM_LENGTH)
        token_mac = cls._get_bits(d, 0, 15)
        computed_mac = cls._securid_shortmac(data[: len(data) - CHECKSUM_LENGTH])
        if token_mac != computed_mac:
            raise InvalidSignature("Invalid checksum")

    @classmethod
    def _numinput_to_bits(cls, data: Bytes, n_bits: int, offset: int = 0) -> bytes:
        bitpos = 13
        out = bytearray(int(MAX_TOKEN_BITS / 8 + 2))
        pos = 0
        for t in data[offset:]:
            decoded = (t - ord("0")) & 0x07
            decoded = decoded << bitpos
            out[0 + pos] = out[0 + pos] | decoded >> 8
            out[1 + pos] = out[1 + pos] | decoded & 0xFF
            bitpos = bitpos - TOKEN_BITS_PER_CHAR
            if bitpos < 0:
                bitpos = bitpos + 8
                pos = pos + 1
            n_bits = n_bits - TOKEN_BITS_PER_CHAR
            if not n_bits:
                break
        return bytes(out)

    @classmethod
    def _bits_to_numoutput(cls, data: Bytes, n_bits: int) -> bytes:
        bitpos = 13
        out = bytearray()
        pos = 0
        for i in range(n_bits, 0, -TOKEN_BITS_PER_CHAR):
            binary = (data[pos] << 8) | data[pos + 1]
            out.append(((binary >> bitpos) & 0x07) + ord("0"))
            bitpos -= TOKEN_BITS_PER_CHAR
            if bitpos < 0:
                bitpos = bitpos + 8
                pos = pos + 1
        return bytes(out)

    @classmethod
    def _get_bits(cls, data: Bytes, start: int, n_bits: int) -> int:
        pos = int(math.floor(start / 8))
        start = start % 8
        val = 0
        for i in range(n_bits, 0, -1):
            val = val << 1
            if (data[pos] << start) & 0x80:
                val = val | 0x01
            start = start + 1
            if start == 8:
                start = 0
                pos = pos + 1
        return val

    @classmethod
    def _set_bits(cls, out: bytearray, start: int, n_bits: int, val: int) -> None:
        pos = int(math.floor(start / 8))
        start = start % 8
        val = val << (32 - n_bits)
        for i in range(n_bits, 0, -1):
            if val & (1 << 31):
                out[pos] = out[pos] | (1 << (7 - start))
            else:
                out[pos] = out[pos] & ~(1 << (7 - start))
            val = val << 1
            start = start + 1
            if start == 8:
                start = 0
                pos = pos + 1

    @classmethod
    def _encrypt_then_xor(cls, key: Bytes, work: bytes) -> bytes:
        out = aes_ecb_encrypt(key, work)
        return bytes([a ^ b for a, b in zip(work, out)])

    @classmethod
    def _securid_mac(cls, data: Bytes) -> bytes:
        # padding
        pad = bytearray(AES_KEY_SIZE)
        p = AES_KEY_SIZE - 1
        i = len(data) * 8
        while i > 0:
            pad[p] = i % 256
            p = p - 1
            i = i >> 8
        # handle the bulk of the input data here
        odd = False
        t = data
        work = bytes([0xFF] * AES_KEY_SIZE)
        while len(t) > AES_KEY_SIZE:
            work = cls._encrypt_then_xor(t[:AES_KEY_SIZE], work)
            t = t[AES_KEY_SIZE:]
            odd = not odd
        # final 0-16 bytes of input data
        work = cls._encrypt_then_xor(t + bytes(AES_KEY_SIZE - len(t)), work)
        # hash an extra block of zeroes, for certain input lengths
        if odd:
            zero = bytearray(AES_KEY_SIZE)
            work = cls._encrypt_then_xor(zero, work)
        # always hash the padding
        work = cls._encrypt_then_xor(pad, work)
        # run hash over current hash value, then return
        return cls._encrypt_then_xor(work, work)

    @classmethod
    def _short_hash(cls, hash_: Bytes) -> int:
        return (hash_[0] << 7) | (hash_[1] >> 1)

    @classmethod
    def _securid_shortmac(cls, data: Bytes) -> int:
        return cls._short_hash(cls._securid_mac(data))

    @classmethod
    def v2_decrypt_seed(cls, enc_seed: Bytes, seed_hash: int) -> bytes:
        key_hash = cls._securid_mac(STOKEN_MAGIC)
        seed = aes_ecb_decrypt(key_hash, enc_seed)
        calc_seed_hash = cls._short_hash(cls._securid_mac(seed))
        if calc_seed_hash != seed_hash:
            raise InvalidSignature("Seed decryption failed")
        return seed

    def get_token(self, password: Optional[str] = None) -> Token:
        """
        Return the Token instance
        """
        return self.token



#!/usr/bin/env python

import argparse
import os
import sys
import time
from datetime import datetime
from typing import List, NoReturn, Optional

from .exceptions import InvalidSignature, ParseException
from .jsontoken import JSONTokenFile
from .sdtid import SdtidFile
from .stoken import DEFAULT_STOKEN_FILENAME, StokenFile
from .token import AbstractTokenFile, Token

__all__ = ["show_token", "export", "interactive", "show_version", "main"]

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_PARSER_ERROR = 2


def show_token(token: Token, pin: Optional[int] = None, verbose: bool = False) -> int:
    "Show token"
    if verbose:
        for_time = datetime.utcnow()
        print(token)
        print("{} {:2}s".format(token.at(for_time, pin=pin), token.time_left(for_time)))
    else:
        print(token.now(pin=pin))
    return EXIT_SUCCESS


def export(token: Token) -> int:
    "Export token into JSON"
    f = JSONTokenFile(token=token)
    sys.stdout.buffer.write(f.export_token())
    sys.stdout.buffer.write(b"\n")
    return EXIT_SUCCESS


def interactive(token: Token, test: bool = False, pin: Optional[int] = None) -> int:
    "Show the code every second until interrupted"
    try:
        while True:
            for_time = datetime.utcnow()
            left = token.time_left(for_time)
            bar = ("#" * left) + ("." * (token.interval - left))
            print("{} {:2}s [{}]".format(token.at(for_time, pin=pin), left, bar))
            if test:
                raise KeyboardInterrupt
            time.sleep(1)
            sys.stdout.write("\033[F\033[K")
    except KeyboardInterrupt:
        pass
    return EXIT_SUCCESS


def show_version(prog: str) -> int:
    "Show version"
    from . import __version__

    print("{} {}".format(prog, __version__))
    return EXIT_SUCCESS


class ArgumentParserException(Exception):
    pass


class ArgumentParser(argparse.ArgumentParser):
    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        if message:
            self._print_message(message, sys.stderr)
        raise ArgumentParserException()


def main(args: Optional[List[str]] = None) -> int:
    prog = os.path.basename(sys.argv[0])
    parser = ArgumentParser()
    parser.add_argument(
        "-f",
        "--filename",
        help="token file (sdtid, stokenrc or json)",
        dest="filename",
        default=DEFAULT_STOKEN_FILENAME,
    )
    parser.add_argument("--password", dest="password")
    parser.add_argument("--pin", type=int, dest="pin")
    parser.add_argument("--export", help="export token into JSON", action="store_true", dest="export")
    parser.add_argument(
        "-i",
        "--interactive",
        help="show the code every second until interrupted",
        action="store_true",
        dest="interactive",
    )
    parser.add_argument("-v", "--verbose", help="verbose output", action="store_true", dest="verbose")
    parser.add_argument(
        "-V",
        "--version",
        help="show version and exit",
        action="store_true",
        dest="version",
    )
    try:
        cli_args = parser.parse_args(args=args)
        if cli_args.version:
            return show_version(prog)
        try:
            f: AbstractTokenFile = JSONTokenFile(cli_args.filename)
        except (FileNotFoundError, ParseException):
            try:
                f = StokenFile(cli_args.filename)
            except (FileNotFoundError, ParseException):
                f = SdtidFile(cli_args.filename)
        token = f.get_token(cli_args.password)
        if cli_args.export:
            return export(token)
        elif cli_args.interactive:
            return interactive(token, pin=cli_args.pin)
        else:
            return show_token(token, pin=cli_args.pin, verbose=cli_args.verbose)
    except ArgumentParserException:
        return EXIT_PARSER_ERROR
    except InvalidSignature:
        if cli_args.password:
            message = "Incorrect password"
        else:
            message = "Error decrypting token - please provide a password"
        print("{}: error: {}".format(prog, message), file=sys.stderr)
        return EXIT_FAILURE
    except Exception as ex:
        print("{}: error: {}".format(prog, ex), file=sys.stderr)
        return EXIT_FAILURE



#!/usr/bin/env python

import json
import os
import os.path
from typing import Any, Dict, Optional, Union

from .exceptions import InvalidSeed, InvalidSerial, ParseException
from .token import SERIAL_LENGTH, AbstractTokenFile, Token
from .utils import Bytes, fromisoformat

__all__ = [
    "JSONTokenFile",
]


class JSONTokenFile(AbstractTokenFile):
    """
    Handler for JSON file format

    Example:

    {
        "digits": 6,
        "exp_date": "2035-12-31",
        "pin": 1234,
        "period": 60,
        "secret": [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25],
        "serial": "000512377827",
        "issuerInt": "myorg",
        "label": "myaccount",
        "type": "SecurID"
    }

    """

    filename: Optional[str]
    token: Token

    def __init__(
        self,
        filename: Optional[str] = None,
        data: Union[bytes, bytearray, str, Dict[str, Any], None] = None,
        token: Optional[Token] = None,
    ) -> None:
        """
        :param filename: JSON file path
        :param data: token as string in JSON format or as a dictionary
        :param token: Token instance
        """
        if token is not None:
            self.filename = None
            self.token = token
        elif data is not None:
            if isinstance(data, str):
                data = bytes(data, "ascii")
            self.filename = None
            self.token = self.json_decode_token(data)
        elif filename is not None:
            self.filename = os.path.expanduser(filename)
            data = self.parse_file(self.filename)
            self.token = self.json_decode_token(data)

    @classmethod
    def parse_file(cls, filename: str) -> bytes:
        """
        Parse JSON file, return content as string

        :param filename: JSON file path
        """
        with open(filename, "rb") as f:
            return f.read()

    @classmethod
    def json_decode_token(cls, data: Union[Bytes, Dict[str, Any]]) -> Token:
        try:
            if isinstance(data, dict):
                dct = data
            else:
                dct = json.loads(data)
            token = Token(
                digits=dct["digits"],
                interval=dct["period"],
                exp_date=fromisoformat(dct["exp_date"]) if dct.get("exp_date") else None,
                seed=bytes(dct["secret"]),
                serial=dct["serial"],
                issuer=dct.get("issuerInt"),
                label=dct.get("label"),
                pin=dct.get("pin"),
            )
            return token
        except json.decoder.JSONDecodeError as ex:
            raise ParseException(ex)

    def get_token(self, password: Optional[str] = None) -> Token:
        """
        Return the Token instance

        :param password: optional password for decrypting the token
        """
        return self.token

    def export_token(self) -> bytes:
        """
        Export token as JSON
        """
        if not self.token.seed:
            raise InvalidSeed("Missing seed")
        if not self.token.serial:
            raise InvalidSerial("Missing serial")
        if len(self.token.serial) != SERIAL_LENGTH:
            raise InvalidSerial("Serial length != {}".format(SERIAL_LENGTH))
        data = {
            "digits": self.token.digits,
            "period": self.token.interval,
            "exp_date": self.token.exp_date.isoformat() if self.token.exp_date else "",
            "secret": [x for x in self.token.seed],
            "serial": self.token.serial,
            "type": "SecurID",
        }
        if self.token.issuer is not None:
            data["issuerInt"] = self.token.issuer
        if self.token.label is not None:
            data["label"] = self.token.label
        if self.token.pin is not None:
            data["pin"] = self.token.pin
        j = json.dumps(data, sort_keys=True)
        return bytes(j, "ascii")



#!/usr/bin/env python

import binascii
import string
from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import Any, Optional, Union

from .exceptions import InvalidSeed
from .utils import (
    AES_KEY_SIZE,
    Bytearray,
    BytesStr,
    aes_ecb_encrypt,
    fromisoformat,
    random,
)

__all__ = ["Token", "AbstractTokenFile", "SERIAL_LENGTH"]

DEFAULT_DIGITS = 6
DEFAULT_INTERVAL = 60
SERIAL_LENGTH = 12
BCD_TIME_BYTES = [2, 3, 4, 5, 8]


class Token(object):
    """
    Handler for RSA SecurID 128-bit compatible token codes.
    """

    serial: str  # serial number
    seed: Optional[bytes]  # decoded AES key
    interval: int  # interval in seconds (30 or 60)
    digits: int  # tokencode digits
    exp_date: Optional[date]  # expiration date
    issuer: Optional[str]  # issuer (origin)
    label: Optional[str]  # label (userlogin, serial)
    pin: Optional[int]  # PIN

    def __init__(
        self,
        serial: BytesStr = "",
        seed: Union[Optional[bytes], Optional[str]] = None,
        interval: int = DEFAULT_INTERVAL,
        digits: int = DEFAULT_DIGITS,
        exp_date: Union[Optional[date], Optional[str]] = None,
        issuer: Optional[str] = None,
        label: Optional[str] = None,
        pin: Optional[int] = 0,
    ) -> None:
        """
        :param serial: token serial number
        :param seed: token seed
        :param interval: time interval in seconds for OTP (default: 60)
        :param digits: number of digits (default: 6)
        :param exp_date: expiration date
        :param issuer: issuer
        :param label: label
        :param pin: PIN (default: 0)
        """
        if not isinstance(serial, str):
            serial = str(serial, "ascii")
        self.serial = serial.zfill(SERIAL_LENGTH)
        if isinstance(seed, str):
            seed = bytes(seed, "ascii")
        if isinstance(exp_date, str):
            exp_date = fromisoformat(exp_date)
        self.seed = seed
        self.interval = interval
        self.digits = digits
        self.exp_date = exp_date
        self.issuer = issuer
        self.label = label
        self.pin = pin

    def generate_otp(self, input: datetime, pin: Optional[int] = None) -> str:
        """
        Generate OTP

        :param input: the time to generate an OTP for
        :returns: OTP code
        """
        if not self.seed:
            raise InvalidSeed("Missing seed")
        key = self.seed
        bcd_time = self._compute_bcd_time(input)
        for bcd_time_bytes in BCD_TIME_BYTES:
            key = aes_ecb_encrypt(key, self._key_from_time(bcd_time, bcd_time_bytes, self.serial))

        return self._token_pin(self._output_code(input, key), pin)

    def at(self, for_time: Union[int, datetime], pin: Optional[int] = None) -> str:
        """
        Generate OTP for the given time
        (accepts either a Unix timestamp integer or a datetime object)

        :param for_time: the time to generate an OTP for
        :returns: OTP code
        """
        if not isinstance(for_time, datetime):
            for_time = datetime.utcfromtimestamp(int(for_time))
        return self.generate_otp(for_time, pin)

    def now(self, pin: Optional[int] = None) -> str:
        """
        Generate the current time OTP

        :returns: OTP value
        """
        return self.generate_otp(datetime.utcnow(), pin)

    def time_left(self, for_time: Union[int, datetime, None] = None) -> int:
        """
        Time until next token

        :returns: seconds
        """
        if for_time is None:
            for_time = datetime.utcnow()
        elif not isinstance(for_time, datetime):
            for_time = datetime.utcfromtimestamp(int(for_time))
        result = (self.interval - for_time.second) % self.interval
        if result == 0:
            result = self.interval
        return result

    def _compute_bcd_time(self, input: datetime) -> bytes:
        """
        Compute BCD time for the given time
        """
        t = input.replace(minute=input.minute & (-2 if self.interval == 30 else -4))
        return binascii.unhexlify(t.strftime("%Y%m%d%H%M0000"))

    def _output_code(self, input: datetime, key: bytes) -> str:
        """
        OTP code output

        :param input: the time to generate an OTP for
        :returns: OTP code
        """
        # key contains 4 consecutive codes
        if self.interval == 30:
            i = ((input.minute & 0x01) << 3) | ((input.second >= 30) << 2)
        else:
            i = (input.minute & 0x03) << 2
        tokencode = (key[i + 0] << 24) | (key[i + 1] << 16) | (key[i + 2] << 8) | key[i + 3]
        return ("0" * self.digits + str(tokencode))[-self.digits :]

    def _token_pin(self, token: str, pin: Optional[int] = None) -> str:
        """
        Support for RSA PIN

        :param token: the generated OTP token
        :param pin: the RSA PIN to integrate
        :returns: OTP code
        """
        pin = pin if pin is not None else self.pin
        if pin is None:
            return token
        else:
            resolved_token = ""
            for i in range(0, len(token)):
                c = int(token[-1])
                token = token[0:-1]
                c += pin % 10
                pin = int(pin / 10)
                resolved_token = "{}{}".format(int(c % 10), resolved_token)
            return resolved_token

    @classmethod
    def _key_from_time(cls, bcd_time: bytes, bcd_time_bytes: int, serial: str) -> bytes:
        key = Bytearray(AES_KEY_SIZE)
        key.arrayset(0xAA, 8)
        key.arraycpy(bcd_time, n=bcd_time_bytes)
        key.arrayset(0xBB, 4, dest_offset=12)
        # write BCD-encoded partial serial number
        for i, p in enumerate(range(4, 12, 2)):
            key[i + 8] = ((ord(serial[p]) - ord("0")) << 4) | (ord(serial[p + 1]) - ord("0"))
        return bytes(key)

    @classmethod
    def random(
        cls,
        serial: BytesStr = "",
        interval: int = DEFAULT_INTERVAL,
        digits: int = DEFAULT_DIGITS,
        exp_date: Optional[date] = None,
        issuer: Optional[str] = None,
        label: Optional[str] = None,
        pin: Optional[int] = None,
    ) -> "Token":
        seed = bytes([random.randint(0, 255) for _ in range(0, AES_KEY_SIZE)])
        """
            Generate a new random token

            :param serial: optional token serial number
            :param interval: time interval in seconds for OTP (default: 60)
            :param digits: number of digits (default: 6)
            :param exp_date: expiration date
            :param issuer: issuer
            :param label: label
            :param pin: PIN
            :returns: the generated Token instance
        """
        if not serial:
            serial = "".join([random.choice(string.digits) for _ in range(0, SERIAL_LENGTH)])
        return Token(
            serial=serial,
            seed=seed,
            interval=interval,
            digits=digits,
            exp_date=exp_date,
            issuer=issuer,
            label=label,
            pin=pin,
        )

    @classmethod
    def _fmt(cls, k: str, v: Any) -> str:
        if v is None:
            return ""
        elif cls.__annotations__[k] in [bytes, Optional[bytes]]:
            return str(binascii.hexlify(v), "ascii")
        else:
            return str(v)

    def __repr__(self) -> str:
        return str(dict([(k, self._fmt(k, v)) for k, v in sorted(self.__dict__.items())]))

    def __str__(self) -> str:
        return " ".join("%s: %s" % (k, self._fmt(k, v)) for k, v in sorted(self.__dict__.items()))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Token) and self.__dict__ == other.__dict__

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


class AbstractTokenFile(ABC):
    """
    Abstract token files handler
    """

    @abstractmethod
    def get_token(self, password: Optional[str] = None) -> Token:  # pragma: no cover
        """
        Return the Token instance

        :param password: optional password for decrypting the token
        """
        raise NotImplementedError



#!/usr/bin/env python

import base64
from collections import OrderedDict
from typing import Any, Dict, Optional
from xml.etree import cElementTree as ET

from .exceptions import InvalidSignature, ParseException
from .token import AbstractTokenFile, Token
from .utils import (
    AES_BLOCK_SIZE,
    AES_KEY_SIZE,
    Bytearray,
    BytesStr,
    aes_ecb_encrypt,
    cbc_hash,
    fromisoformat,
    xor_block,
)

__all__ = [
    "SdtidFile",
]

TOKEN_ENC_IV = bytes(
    [
        0x16,
        0xA0,
        0x9E,
        0x66,
        0x7F,
        0x3B,
        0xCC,
        0x90,
        0x8B,
        0x2F,
        0xB1,
        0x36,
        0x6E,
        0xA9,
        0x57,
        0xD3,
    ]
)
BATCH_MAC_IV = bytes(
    [
        0x2B,
        0x7E,
        0x15,
        0x16,
        0x28,
        0xAE,
        0xD2,
        0xA6,
        0xAB,
        0xF7,
        0x15,
        0x88,
        0x09,
        0xCF,
        0x4F,
        0x3C,
    ]
)
TOKEN_MAC_IV = bytes(
    [
        0x1B,
        0xB6,
        0x7A,
        0xE8,
        0x58,
        0x4C,
        0xAA,
        0x73,
        0xB2,
        0x57,
        0x42,
        0xD7,
        0x07,
        0x8B,
        0x83,
        0xB8,
    ]
)
MAX_HASH_DATA = 65536


class SdtidFile(AbstractTokenFile):
    """
    Handler for RSA SecurID sdtid XML file format.
    """

    filename: str
    values: Dict[str, Any]  # sdtid file as OrderedDict
    token: Token

    def __init__(self, filename: str) -> None:
        self.filename = filename
        self.parse_file(filename)

    def parse_file(self, filename: str) -> None:
        """
        Parse sdtid file

        :param filename: sdtid file path
        """
        try:
            with open(filename, "r") as f:
                xml = ET.XML(f.read())
        except ET.ParseError:
            raise ParseException("Error parsing {}".format(filename))
        self.values = self.xml_to_dict(xml)
        serial = self.get("SN")
        interval = self.get("Interval", default=60, kind="int")
        digits = self.get("Digits", default=6, kind="int")
        exp_date = self.get("Death", kind="date")
        issuer = self.get("Origin")
        label = self.get("UserLogin") or serial
        self.token = Token(
            serial=serial,
            interval=interval,
            digits=digits,
            exp_date=exp_date,
            issuer=issuer,
            label=label,
        )

    def get_token(self, password: Optional[str] = None) -> Token:
        """
        Return the Token instance

        :param password: optional password for decrypting the token
        """
        if self.token.seed is None:
            self.token.seed = self.decrypt_seed(password)
        return self.token

    def decrypt_seed(self, password: Optional[str] = None) -> bytes:
        secret = self.get("Secret", kind="base64")
        dest = self.get("Dest")
        name = self.get("Name")
        if password is None:
            password = self.get("Origin")
        key = self.decrypt_secret(secret, name, self.hash_password(password, dest, name))
        self.verify_mac(
            "Header",
            "BatchMAC",
            self.get("Name"),
            "HeaderMAC",
            "TKNHeader",
            key,
            BATCH_MAC_IV,
        )
        self.verify_mac("Token", "TokenMAC", self.get("SN"), "TokenMAC", "TKN", key, TOKEN_MAC_IV)
        enc_seed = self.get("Seed", kind="base64")
        token_enc_key = self.calc_key("TokenEncrypt", self.token.serial, key, TOKEN_ENC_IV)
        return self.decrypt_enc_seed(enc_seed, self.token.serial, token_enc_key)

    def verify_mac(
        self,
        kind: str,
        str0: str,
        str1: str,
        node: str,
        section: str,
        key1: bytes,
        iv: bytes,
    ) -> None:
        mac_key = self.calc_key(str0, str1, key1, iv)
        mac = self.get(node, kind="base64")
        hs = SessionHash()
        hs.recursive_hash(section, self.values["TKNBatch"][section])
        mac_calc = hs.compute_hash(mac_key, iv)
        if mac != mac_calc:
            raise InvalidSignature("{} MAC check failed".format(kind))

    @classmethod
    def hash_password(cls, password: str, salt0: str, salt1: str) -> bytes:
        key = Bytearray(AES_KEY_SIZE)
        key.arraycpy(salt1)

        data = Bytearray(0x50)
        data.arraycpy(password, n=0x20)
        data.arraycpy(salt0, n=0x20, dest_offset=0x20)

        result = bytes(AES_KEY_SIZE)
        iv = bytes(AES_BLOCK_SIZE)
        for i in range(0, 1000):
            data[0x4F] = (i >> 0) % 256
            data[0x4E] = (i >> 8) % 256
            result = xor_block(result, cbc_hash(key, iv, data))
        return bytes(result)

    @classmethod
    def decrypt_secret(cls, enc_bin: bytes, str0: bytes, key: bytes) -> bytes:
        buf = Bytearray(AES_KEY_SIZE)
        buf.arraycpy("Secret", n=8)
        buf.arraycpy(str0, n=8, dest_offset=8)
        return xor_block(aes_ecb_encrypt(key, buf), enc_bin)

    @classmethod
    def decrypt_enc_seed(cls, enc_bin: bytes, str0: str, key: bytes) -> bytes:
        buf = Bytearray(AES_KEY_SIZE)
        buf.arraycpy(str0, n=8)
        buf.arraycpy("Seed", n=8, dest_offset=8)
        return xor_block(aes_ecb_encrypt(key, buf), enc_bin)

    @classmethod
    def calc_key(cls, str0: str, str1: str, key: bytes, iv: bytes) -> bytes:
        buf = Bytearray(64)
        buf.arraycpy(str0, n=32)
        buf.arraycpy(str1, n=32, dest_offset=32)
        return cbc_hash(key, iv, buf)

    @classmethod
    def xml_to_dict(cls, xml: ET.Element) -> Dict[str, Any]:
        """
        Convert XML to nested OrderDict
        """
        if xml:
            dd: Dict[str, Any] = OrderedDict()
            for dc in map(cls.xml_to_dict, list(xml)):
                for k, v in dc.items():
                    if k in dd:
                        if not isinstance(dd[k], list):
                            dd[k] = [dd[k]]
                        dd[k].append(v)
                    else:
                        dd[k] = v
            return OrderedDict({xml.tag: dd})
        else:
            return OrderedDict({xml.tag: (xml.text or "")})

    def get(self, name: str, default: Any = None, kind: Optional[str] = None) -> Any:
        value = self.values["TKNBatch"]["TKN"].get(name)
        if value is None:
            value = self.values["TKNBatch"]["TKNHeader"].get(name)
        if value is None:
            value = self.values["TKNBatch"]["TKNHeader"].get("Def" + name)
        if value is None:
            value = default
        if kind == "base64":
            if value[0] == "=":
                value = value[1:]
            value = base64.b64decode(value)
        elif kind == "int":
            value = int(value)
        elif kind == "date":
            value = fromisoformat(value.replace("/", "-"))
        return value


class SessionHash(object):
    pos: int = 0
    padding: int = 0
    data: Bytearray

    def __init__(self) -> None:
        self.data = Bytearray(MAX_HASH_DATA)

    def recursive_hash(self, name: str, node: Dict[str, Any]) -> None:
        # from https://github.com/cernekee/stoken
        for k, v in node.items():
            if k.endswith("MAC"):
                continue
            longname = "%s.%s" % (name, k)
            remain = MAX_HASH_DATA - self.pos
            if isinstance(v, dict):
                self.recursive_hash(longname, v)
            else:
                if not v:
                    # "An empty string is valid XML but it might violate
                    # the sdtid format.  We'll handle it the same bizarre
                    # way as RSA just to be safe."
                    data = "%s </%s>\n" % (longname, name)
                    self._append_data(data)
                else:
                    data = "%s %s\n" % (longname, v)
                    self._append_data(data)
                    length = len(data) + self.padding
                    if length <= 16 and length < remain:
                        self.pos = self.pos & ~0xF
                        self._append_data(data, n=min(len(data), remain))
                        self.data.arrayset(0, dest_offset=self.pos + len(data), n=self.padding)
                #  This doesn't really make sense but it's required for compatibility
                self.pos = self.pos + len(data) + self.padding
                self.padding = self.pos & 0xF or 0x10

    def _append_data(self, data: BytesStr, n: Optional[int] = None) -> None:
        if isinstance(data, str):
            data = bytes(data, "ascii")
        if n is None:
            n = len(data)
        self.data[self.pos : self.pos + n] = data[:n]

    def compute_hash(self, key: bytes, iv: bytes) -> bytes:
        return cbc_hash(key, iv, self.data[0 : self.pos])



#!/usr/bin/env python

__all__ = [
    "ParseException",
    "InvalidToken",
    "InvalidSignature",
    "InvalidSeed",
    "InvalidSerial",
]


class ParseException(Exception):
    """
    This is raised in case of error parsing file
    """

    pass


class InvalidToken(Exception):
    """
    This is raised in case of invalid token
    """

    pass


class InvalidSignature(Exception):
    """
    This is raised when signature verification fails.
    This can occur when password is required for decrypting the token.
    """


class InvalidSeed(Exception):
    """
    This is raised when the seed is missing or invalid.
    """


class InvalidSerial(Exception):
    """
    This is raised when the serial is missing or invalid.
    """
